# include	"FingerEngine.h"
# include	<CThread.h>
# include	"CPlanWindow.h"
# include	<CTCPStream.h>
# include	<StBind.h>
# include	"CFingerApplication.h" // XXX - For database defines

# define	BUFFER_SIZE		(1024 * 10)

enum {
	ERR_CANT_PARSE = 100
};

//=========================================================================
//	* Address Manipulation
//=========================================================================

ErrorCode ParseFingerAddress( const string& inAddress, string& outUserName,
	string& outHostName )
{
	ErrorCode err = noErr;

		// Scan for '@'
	int i;
	for ( i = 0; inAddress[i] != NULL && inAddress[i] != '@'; i++ )
		;
	RETURN_ON( i == inAddress.length( ), ERR_CANT_PARSE );

	outUserName = inAddress.substring( 0, i );
	outHostName = inAddress.substring( i + 1, inAddress.length( ) - (i + 1) );
	RETURN_ON( outHostName == string( "" ), ERR_CANT_PARSE );

clean_up:
	return err;
}


//=========================================================================
//	* Plan display
//=========================================================================

class CDisplayPlanThread : public CThread
{
public:
					CDisplayPlanThread(
						const char* inUserName,
						const char* inHostName );
private:
	string			mUserName;
	string			mHostName;

	virtual void	RunThread( );
	virtual void	ThreadFinished( );
};

void DisplayPlan( const char* inAddress )
{
	string userName, hostName;
	ErrorCode err = ParseFingerAddress( inAddress, userName, hostName );
	if ( err != noErr )
	{
		BAlert* alert = new BAlert( "", string("The address \"") + inAddress
			+ "\" is not a valid Finger adress.", "OK", NULL, NULL,
			B_WIDTH_AS_USUAL, B_STOP_ALERT );
		alert->Go( );
	}
	DisplayPlan( userName, hostName );
}

void DisplayPlan( const char* inUserName, const char* inHostName )
{
	CThread* thread = new CDisplayPlanThread( inUserName, inHostName );
	thread->Resume( );
}

CDisplayPlanThread::CDisplayPlanThread(
	const char* inUserName, const char* inHostName )
:	CThread( string("Fingering ") + inUserName + "@" + inHostName,
	B_NORMAL_PRIORITY ), mUserName( inUserName ), mHostName( inHostName )
{
	be_app->PostMessage( MSG_ADD_CLIENT );
}

void CDisplayPlanThread::ThreadFinished( )
{
	be_app->PostMessage( MSG_SUBTRACT_CLIENT );
}

void CDisplayPlanThread::RunThread( )
{
	string realName;
	StBind<CGrowBlock> plan = FetchPlan( mUserName, mHostName, realName );
	
	if ( plan.ptr( ) )
		new CPlanWindow( (char*) plan->Data( ), plan->DataSize( ),
			mUserName, mHostName, realName ); 
}


//=========================================================================
//	* Plan fetching
//=========================================================================

static void UpdateDatabase( const char* inUserName, const char* inHostName,
	string& outRealName );

CGrowBlock* FetchPlan( const char* inUserName, const char* inHostName,
	string& outRealName )
{
	ErrorCode err = noErr;

	StBind<CGrowBlock> plan = new CGrowBlock( string("Plan of ")
		+ inUserName + "@" + inHostName );

	CTCPStream stream( "Finger client stream" );
	RETURN_ON_ERR( stream.Error( ) );
	stream.Connect( inHostName, 79 );
	RETURN_ON_ERR( stream.Error( ) );
	
	stream.WriteLine( inUserName );
	
		// Receive server response
	while ( TRUE )
	{
		char buffer[BUFFER_SIZE];
		int count = stream.Read( buffer, BUFFER_SIZE );

		if ( count < 0 )
		{
			stream.Close( );
			if ( stream.Error( ) == ENOTCONN )
				break;
			else
				RETURN( stream.Error( ) );
		}

		plan->Append( buffer, count );	 
	}

		// Strip extra carriage returns
	char* text	= (char*) plan->Data( );
	int len		= plan->DataSize( );
	int deleted	= 0;
	for ( int i = 0; i < len; i++ )
	{
		if ( text[i] == '\r' )
			deleted++;
		else
			text[i - deleted] = text[i];
	}
	plan->SetSize( len - deleted );
	
	UpdateDatabase( inUserName, inHostName, outRealName );

clean_up:
	if ( err )
	{
		ReportError( err, string("Could not finger \"") + inUserName
			+ "@" + inHostName + "\"." );
		return NULL;
	}
	return plan.release( );
}

void UpdateDatabase( const char* inUserName, const char* inHostName,
	string& outRealName )
{
	BDatabase* db = boot_volume( ).Database( );
	BTable* fingerTable = db->FindTable( FINGERABLE_USER_TABLE );
	
	double currentTime = time( NULL );
	
	StBind<BQuery> query = new BQuery( );

	query->AddTable( fingerTable );
	query->PushField( "User Name" );
	query->PushString( inUserName );
	query->PushOp( B_EQ );
	query->PushField( "Host Name" );
	query->PushString( inHostName );
	query->PushOp( B_EQ );
	query->PushOp( B_AND );
	
	query->FetchOne( );
	
	StBind<BRecord> rec;
	if ( query->CountRecordIDs( ) > 0 )
	{
		rec = new BRecord( db, query->RecordIDAt( 0 ) );
		int count = rec->FindLong( "Count" );
		rec->SetLong( "Count", count + 1 );
		
			// Grab "official" real name from database.
		outRealName = rec->FindString( "Name" );
	}
	else
	{
		outRealName = string(inUserName) + "@" + inHostName;

		rec = new BRecord( fingerTable );
		rec->SetString( "Name", outRealName );
		rec->SetTime( "Created", currentTime );
		rec->SetTime( "Modified", currentTime );
		rec->SetString( "User Name", inUserName );
		rec->SetString( "Host Name", inHostName );
		rec->SetLong( "Count", 1 );
	}
	
	rec->SetLong( "fsType", 'Fing' );
	rec->SetLong( "fsCreator", 'Fing' );
	rec->SetTime( "Last Fingered", currentTime );
	rec->Commit( );
}