
# include	"CFingerApplication.h"
# include	"FingerEngine.h"
# include	"CFingerPromptWindow.h"
# include	<stringclass.h>

static const char* URL_SCHEMES[] = { "finger", NULL };

void main( )
{
	CFingerApplication* theApp = new CFingerApplication;
	theApp->Run( );
	delete theApp;
}

CFingerApplication::CFingerApplication( )
	: CStandardApplication( 'Fing' )
{
	InitMainMenu( );
	InitFingerTable( );
	mReceivedValidRefs = FALSE;
	
	mIC = new ICInstance( 'Fing' );
	mIC->RegisterURLSchemes( URL_SCHEMES );
}

CFingerApplication::~CFingerApplication( )
{
	delete mIC;
}

void CFingerApplication::InitMainMenu( )
{
	BMenuItem* item;
	BPopUpMenu* menu = new BPopUpMenu( "", FALSE, FALSE );
	
	item = new BMenuItem( "About Finger...",
		new BMessage( B_ABOUT_REQUESTED ) );
	item->SetTarget( this );
	menu->AddItem( item );

	menu->AddSeparatorItem( );

	item = new BMenuItem( "Finger User",
		new BMessage( MSG_NEW_FINGER ), 'F' );
	item->SetTarget( this );
	menu->AddItem( item );

	item = new BMenuItem( "Test IC_DISPLAY_URL",
		new BMessage( MSG_TEST_DISPLAY_URL ) );
	item->SetTarget( this );
	menu->AddItem( item );

	menu->AddSeparatorItem( );
		
	item = new BMenuItem( "Quit",
		new BMessage( B_QUIT_REQUESTED ), 'Q' );
	item->SetTarget( this );
	menu->AddItem( item );

	SetMainMenu( menu );
}

void CFingerApplication::InitFingerTable( )
{
	BDatabase* db = boot_volume( ).Database( );
	BTable* fingerTable;

	db->Sync( );

	fingerTable = db->FindTable( FINGERABLE_USER_TABLE );
	if ( !fingerTable )
		fingerTable = db->CreateTable( FINGERABLE_USER_TABLE, "BrowserItem" );
		
	if ( fingerTable->FieldKey( "User Name" ) == B_ERROR )
		fingerTable->AddStringField( "User Name", B_INDEXED_FIELD );
	if ( fingerTable->FieldKey( "Host Name" ) == B_ERROR )
		fingerTable->AddStringField( "Host Name", B_INDEXED_FIELD );
	if ( fingerTable->FieldKey( "Last Fingered" ) == B_ERROR )
		fingerTable->AddTimeField( "Last Fingered" );
	if ( fingerTable->FieldKey( "Count" ) == B_ERROR )
		fingerTable->AddLongField( "Count" );

	fingerTable->SetTypeAndApp( 'Fing', 'Fing' );	
}

void CFingerApplication::MessageReceived( BMessage* inMessage )
{
	switch ( inMessage->what )
	{
		case MSG_NEW_FINGER:
			NewFinger( );
			break;
		
		case MSG_TEST_DISPLAY_URL:
			TestDisplayURL( );
			break;		

		case IC_DISPLAY_URL:
			DisplayURL( inMessage->FindString( "url" ) );
			break;
		
		case B_RECORD_ADDED:
		case B_RECORD_REMOVED:
		case B_RECORD_MODIFIED:
			update_query( inMessage );
			break;
		
		default:
			CStandardApplication::MessageReceived( inMessage );
	}
}

void CFingerApplication::ArgvReceived( int argc, char** argv )
{
	for ( int i = 1; i < argc; i++ )
		DisplayPlan( argv[i] );
	mReceivedValidRefs = TRUE;
}

void CFingerApplication::AboutRequested( )
{
	BAlert* about = new BAlert( "", "Finger 1.0b3\n\n" \
		"Copyright 1996 Eric Kidd <eric.kidd@pobox.com>\n" \
		"Icon with help from Neil Miller", "OK" );
	about->Go( );
}

void CFingerApplication::RefsReceived( BMessage* inMessage )
{
	ulong type;
	long count;
	
	inMessage->GetInfo( "refs", &type, &count );
	for ( int i = 0; i < count; i++ )
	{
		record_ref item = inMessage->FindRef( "refs", i );
		if ( item.record >= 0 && item.database >= 0 )
		{
			if ( does_ref_conform( item, FINGERABLE_USER_TABLE ) )
			{
				BRecord* rec = new BRecord( item );
				if ( rec->Error( ) == B_NO_ERROR )
				{
					string user = rec->FindString( "User Name" );
					string host = rec->FindString( "Host Name" );
					DisplayPlan( user, host );
					mReceivedValidRefs = TRUE;	
				}
			}
		}
	}
}

void CFingerApplication::ReadyToRun( )
{
	if ( !mReceivedValidRefs )
		PostMessage( MSG_NEW_FINGER );
}	

void CFingerApplication::NewFinger( )
{
	new CFingerPromptWindow( );
}

void CFingerApplication::TestDisplayURL( )
{
	mIC->DisplayURL( "finger:@mit.edu" );
}

void CFingerApplication::DisplayURL( const char* inURL )
{
	string url = inURL;
	int split = url.findchar( ':' );
	
	if ( split == -1 )
	{
		BAlert* alert = new BAlert( "", "What gave you the idea that <"
			+ url + "> was a valid URL?", "OK", NULL, NULL, B_WIDTH_AS_USUAL,
			B_STOP_ALERT );
		alert->Go( );
	}
	
	DisplayPlan( url.tailstring( split + 1 ) );
}