# include	"CPlanWindow.h"
# include	"CFingerApplication.h" // XXX - For database defines
# include	"FingerEngine.h"
# include	<CTCPStream.h>
# include	<StBind.h>

# include	<stdio.h>
# include	<string.h>
# include	<errno.h>

# define	SCROLL_WIDTH	(13.0)
# define	TEXT_INSET		(4.0)
# define	MIN_PLAN_WIDTH	(300.0)
# define	MIN_PLAN_HEIGHT	(50.0)
# define	H_PLAN_PAD		(SCROLL_WIDTH + TEXT_INSET * 2)
# define	V_PLAN_PAD		(SCROLL_WIDTH + TEXT_INSET * 2) // + MENU_HEIGHT

	// XXX - SetSizeLimits seems to be broken in DR7.
	// XXX - SetSizeLimits fixed in DR8, Zoom( ) now broken.
# define	H_FUDGE			(8)
# define	V_FUDGE			(27)

enum /* Messages */ {
	MSG_INIT_WINDOW			= 'iwin',
	MSG_SAVE_COPY_AS		= 'sca_',
	MSG_REALLY_SAVE_COPY_AS	= 'sca!',
	MSG_COPY_ALL			= 'call',
	MSG_REFINGER			= 'rfng',
	MSG_FINGER_HOST			= 'fhst',
	MSG_FINGER_OTHER		= 'foac'
};

CStaggeredWindowArranger CPlanWindow::mArranger(
	BPoint( 80, 30 ),		// Origin
	BPoint( 600, 200 ),		// Window size
	BPoint( 10, 20 ),		// Per-window offset
	6 );					// Max windows before restarting

CPlanWindow::CPlanWindow(
	const char*	inPlan,
	int			inPlanLength,
	const char* inUserName,
	const char* inHostName,
	const char* inRealName )
: 	CStandardWindow( mArranger.NextFrame( ), "Plan Window",
		B_DOCUMENT_WINDOW, B_WILL_ACCEPT_FIRST_CLICK ),
	mUser(inUserName), mHost(inHostName), mRealName(inRealName)
{
	BRect rect;
	
	Lock( );

		// Add menu bar to window
	rect.top = 0;
	rect.left = 0;
	mMenuBar = new BMenuBar( rect, "Plan" );
	InitializeMenus( );
	AddChild( mMenuBar );
	
		// Find out how big our menu bar is.
	rect = mMenuBar->Bounds( );
	mMenuHeight = rect.bottom;

		// Set up scrolling text area for plan	
	rect = Frame( );
	rect.OffsetTo( BPoint( -1, -1 ) );
	rect.top = mMenuHeight + 1; // XXX - +1 works around DR8 display artifact
	rect.bottom -= SCROLL_WIDTH;
	rect.right -= SCROLL_WIDTH;
	mPlanView = new BTextView( rect, "Plan text", BRect( TEXT_INSET,
		TEXT_INSET, 5, 5 ), B_FOLLOW_ALL_SIDES, B_WILL_DRAW );
	
	mScrollView = new BScrollView( "Scroll", mPlanView, B_FOLLOW_ALL_SIDES,
		B_WILL_DRAW, TRUE, TRUE, FALSE );
	AddChild( mScrollView );
	
	mPlanView->SetFontName( "Kate" );
	mPlanView->SetFontSize( 9 );
	mPlanView->SetAlignment( B_ALIGN_LEFT );
	mPlanView->MakeSelectable( FALSE );
	mPlanView->MakeEditable( FALSE );
	mPlanView->SetAutoindent( FALSE );
	mPlanView->SetTabWidth( 7 * 4 );

		// XXX - Should adjust big steps dynamically
	mScrollView->ScrollBar( B_HORIZONTAL )->SetSteps( 12, 120 );
	mScrollView->ScrollBar( B_VERTICAL )->SetSteps( 12, 120 );

	mPlanView->Insert( inPlan, inPlanLength );
	RecalculateWindowSize( );

		// Set window title appropriately
	string address = mUser + "@" + mHost;
	if ( address != mRealName )
		SetTitle( mRealName + " (" + address + ")" );
	else 
		SetTitle( address );
		
	Unlock( );
	
	Show( );
}

CPlanWindow::~CPlanWindow( )
{
	delete mShortcutMenu;
}

void CPlanWindow::MessageReceived( BMessage* inMessage )
{
	switch ( inMessage->what )
	{
		case MSG_SAVE_COPY_AS:
			SaveACopyAs( );
			break;

		case MSG_REALLY_SAVE_COPY_AS:
			SaveACopyAsRequested( inMessage );
			break;
			
		case MSG_COPY_ALL:
			CopyAll( );
			break;
			
		case MSG_REFINGER:
			DisplayPlan( mUser, mHost );
			break;

		case MSG_FINGER_HOST:
			DisplayPlan( "", mHost );
			break;	

		case MSG_FINGER_OTHER:
			FingerOther( );
			break;
			
		default:
			CStandardWindow::MessageReceived( inMessage );
	}
}

bool CPlanWindow::FilterMouseDown( BPoint inWhere, BView** inView )
{
		// Since we set B_WILL_ACCEPT_FIRST_CLICK,
		// bring ourselves to the front when needed.
	if ( !IsActive( ) )
	{
		Activate( );
		UpdateIfNeeded( );
	}

	if ( *inView != mPlanView )
		return TRUE;
	
	long buttons = CurrentMessage( )->FindLong( "buttons" );
	if ( buttons != B_SECONDARY_MOUSE_BUTTON )
		return TRUE;
	
	mPlanView->ConvertToScreen( &inWhere );
	BMenuItem* item = mShortcutMenu->Go( inWhere, FALSE );
	UpdateIfNeeded( );
	if ( item )
		PostMessage( new BMessage( item->Message( ) ) );
	return FALSE;
}


void CPlanWindow::RecalculateWindowSize( )
{
		// Get line widths
	long lineCount = mPlanView->CountLines( );
	float maxWidth = MIN_PLAN_WIDTH;
	for ( long i = 0; i < lineCount; i++ )
	{
		float lineWidth = mPlanView->LineWidth( i );
		if ( lineWidth > maxWidth )
			maxWidth = lineWidth;
	}

		// Get total height
	font_info fi;
	mPlanView->GetFontInfo( &fi );
	float lineHeight = fi.ascent + fi.descent + fi.leading;
	float totalHeight = lineHeight * lineCount;
	if ( totalHeight < MIN_PLAN_HEIGHT )
		totalHeight = MIN_PLAN_HEIGHT;

		// Resize window	
	BRect windowBounds = Frame( );
	ConvertFromScreen( &windowBounds );
	windowBounds.right = maxWidth + H_PLAN_PAD;
	windowBounds.bottom = totalHeight + V_PLAN_PAD + mMenuHeight;
	ResizeTo( windowBounds.right, windowBounds.bottom );
	
		// Set minimum and maximum window sizes
		// XXX - Maximum is padded because of Zoom( ) bug.
		// Zoom has not been fixed to take the fix to
		// SetSizeLimits( ) into account, and assumes that
		// the latter function still includes border sizes.
		// XXX - Need to fix minimum for scrollbar display.
	int minWWidth, maxWWidth, minWHeight, maxWHeight;
	minWWidth	= MIN_PLAN_WIDTH / 2 + H_PLAN_PAD;
	maxWWidth	= windowBounds.right + H_FUDGE;
	minWHeight	= MIN_PLAN_HEIGHT / 2 + V_PLAN_PAD + mMenuHeight;
	maxWHeight	= windowBounds.bottom + V_FUDGE;
	SetSizeLimits( minWWidth, maxWWidth, minWHeight, maxWHeight );
	SetZoomLimits( maxWWidth, maxWHeight );
	
		// Resize window to fit on screen
		// XXX - Leaves window zoomed. Hmm.
	Zoom( );
	
		// XXX - Didn't work in DR7 or DR8
//	BRect textRect = mPlanView->TextRect( );
//	textRect.right = maxWidth + TEXT_INSET;
//	mPlanView->SetTextRect( textRect );
}

void CPlanWindow::InitializeMenus( )
{
		//===== File Menu

	BMenu* fileMenu = new BMenu( "File" );

	BMenuItem* saveCopyAsItem = new BMenuItem( "Save a Copy As...",
		new BMessage( MSG_SAVE_COPY_AS ), 'S' );
	fileMenu->AddItem( saveCopyAsItem );

	fileMenu->AddItem( new BMenuItem( "Close",
		new BMessage( B_CLOSE_REQUESTED ), 'W' ) );

	mMenuBar->AddItem( fileMenu );

		//===== Edit Menu

	BMenu* editMenu = new BMenu( "Edit" );

	BMenuItem* copyItem = new BMenuItem( "Copy", NULL, 'C' );
	copyItem->SetEnabled( FALSE );
	editMenu->AddItem( copyItem );

	BMenuItem* copyAllItem = new BMenuItem( "Copy All",
		new BMessage( MSG_COPY_ALL ) );
	editMenu->AddItem( copyAllItem );

	BMenuItem* selectAllItem = new BMenuItem( "Select All", NULL, 'A' );
	selectAllItem->SetEnabled( FALSE );
	editMenu->AddItem( selectAllItem );

	mMenuBar->AddItem( editMenu );

		//===== Finger menu

	BMenu* fingerMenu = new BMenu( "Finger" );
	
	BMenuItem* fingerAgainItem = new BMenuItem( "Finger Again",
		new BMessage( MSG_REFINGER ), 'R' );
	fingerMenu->AddItem( fingerAgainItem );
		
	BMenuItem* fingerHostItem = new BMenuItem( "Finger Host",
		new BMessage( MSG_FINGER_HOST ) );
	fingerMenu->AddItem( fingerHostItem );
	
	BMenuItem* fingerOtherItem = new BMenuItem( "Finger Other Accounts",
		new BMessage( MSG_FINGER_OTHER ) );
	fingerMenu->AddItem( fingerOtherItem );
	
	mMenuBar->AddItem( fingerMenu );

		//===== Popup menu
		// XXX - Code duplication

	BMenuItem* saveCopyAsItem2 = new BMenuItem( "Save a Copy As...",
		new BMessage( MSG_SAVE_COPY_AS ), 'S' );
	BMenuItem* copyAllItem2 = new BMenuItem( "Copy All",
		new BMessage( MSG_COPY_ALL ) );
	BMenuItem* fingerAgainItem2 = new BMenuItem( "Finger Again",
		new BMessage( MSG_REFINGER ), 'R' );
	BMenuItem* fingerHostItem2 = new BMenuItem( "Finger Host",
		new BMessage( MSG_FINGER_HOST ) );
	BMenuItem* fingerOtherItem2 = new BMenuItem( "Finger Other Accounts",
		new BMessage( MSG_FINGER_OTHER ) );

	mShortcutMenu = new BPopUpMenu( "Plan Shortcuts", FALSE, FALSE );
	mShortcutMenu->AddItem( copyAllItem2 );
	mShortcutMenu->AddSeparatorItem( );
	mShortcutMenu->AddItem( fingerAgainItem2 );
	mShortcutMenu->AddItem( fingerHostItem2 );
	mShortcutMenu->AddItem( fingerOtherItem2 );
	mShortcutMenu->AddSeparatorItem( );
	mShortcutMenu->AddItem( saveCopyAsItem2 );

		// Disable certain items if we're fingering a host
		// rather than a user
	if ( mUser == "" )
	{
		fingerHostItem->SetEnabled( FALSE );
		fingerOtherItem->SetEnabled( FALSE );
		fingerHostItem2->SetEnabled( FALSE );
		fingerOtherItem2->SetEnabled( FALSE );
	}
}

void CPlanWindow::CopyAll( )
{
	string plan( mPlanView->Text( ), mPlanView->TextLength( ) );

	be_clipboard->Lock( );
	be_clipboard->Clear( );
	be_clipboard->AddText( plan );
	be_clipboard->Commit( );
	be_clipboard->Unlock( );	
}

void CPlanWindow::SaveACopyAs( )
{
	string name = Title( );
	name.truncate( B_FILE_NAME_LENGTH );
	RunSavePanel( name, "Finger: Save a Copy of \"" + name + "\" As...",
		NULL, NULL, new BMessage( MSG_REALLY_SAVE_COPY_AS ) );
}

void CPlanWindow::SaveACopyAsRequested( BMessage* inMessage )
{
	ErrorCode err = noErr;
	
		// Find our directory
	record_ref directoryRef = inMessage->FindRef( "directory" );
	StBind<BDirectory> directory = new BDirectory( directoryRef );
	RETURN_ON_ERR( directory->Error( ) );
	
		// Create our file
	StBind<BFile> file = new BFile( );
	RETURN_ON_ERR( directory->Create( inMessage->FindString( "name" ), file.ptr() ) );
	
		// Set file type and write out plan
	RETURN_ON_ERR( file->Open( B_READ_WRITE ) );
	RETURN_ON_NEGATIVE( file->Write( mPlanView->Text( ),
		mPlanView->TextLength( ) ) );
	file->SetTypeAndApp( 'TEXT', 'EDIT' );

		// XXX - This shouldn't be necessary, but is.
		// Is this my bug or a DR8 bug?
	file->Close( );

clean_up:	
	if ( err != B_NO_ERROR )
	{
		BAlert* alert = new BAlert( "", string("Finger could not save the file \"")
			+ inMessage->FindString( "name" ) + "\".", "OK", NULL, NULL,
			B_WIDTH_AS_USUAL, B_STOP_ALERT );
		alert->Go( );
	}	
}

void CPlanWindow::FingerOther( )
{
	BDatabase* db = boot_volume( ).Database( );
	BTable* fingerTable = db->FindTable( FINGERABLE_USER_TABLE );
	StBind<BQuery> query = new BQuery( );
	
		// same user or real names...
	query->AddTable( fingerTable );
	query->PushField( "Name" );
	query->PushString( mRealName );
	query->PushOp( B_EQ );
	query->PushField( "User Name" );
	query->PushString( mUser );
	query->PushOp( B_EQ );
	query->PushOp( B_OR );
	
		// ...but not same account
	query->PushField( "User Name" );
	query->PushString( mUser );
	query->PushOp( B_NE );
	query->PushField( "Host Name" );
	query->PushString( mHost );
	query->PushOp( B_NE );
	query->PushOp( B_OR );

	query->PushOp( B_AND );
	
	query->Fetch( );
	
	long count = query->CountRecordIDs( );
	if ( count > 0 )
	{
		for ( long i = 0; i < count; i++ )
		{
			StBind<BRecord> rec = new BRecord( db, query->RecordIDAt( i ) );
			DisplayPlan( rec->FindString( "User Name" ),
				rec->FindString( "Host Name" ) );
		}
	}
	else
	{
		BAlert* alert = new BAlert( "", "Finger can't find any other accounts " \
			"belonging to this user. Try fingering this user's other " \
			"accounts manually.", "OK" );
		alert->Go( );
	}
}
